# This file contains the TemporalSelector base class
# The TemporalSelector object responds to graph interface communication
# received through its recv_graph_comm method and sends graph
# interface communication through the send_graph_comm method.
# 

# Object fields
# 
# input_id_list_ : list of input ids. The same ids are expected to be used
#                  by individual execution subgraphs
# input_info_ : array of info on input_ids. For each input id $i,
#       ($i,spec) : multicast spec for input data
#       ($i,trigger) : 0/1 indicating if input is execution trigger
#       ($i,selector) : selector object for this input
# vagent_array_ : array associated by "addr,port" indexes that holds
#                 the name of video agents used by this object.
# subgraph_list_ : list of subgraph identifiers. These identifiers are
#                  local only.
# subgraph_info_ : array of info on subgraphs. For each subgraph $s and
#                  each input id $i,
#       ($s,comm_obj) : graph comm object for talking to this subgraph
#       ($s,spec) : cntrl spec for this subgraph
# id_ : id of this object to be used in graph communication
# comm_obj_ : object for graph comm (from above, NOT to subgraphs)
# callback_array_ : holds comm callbacks with index <sid>,<mid>

import AnnounceListenManager/GraphComm
import PsvpVideoAgent
import TemporalSelectorInput

Class TemporalSelector

TemporalSelector instproc init {id control_spec} {
    $self next;

    $self instvar input_id_list_;
    $self instvar input_info_;
    $self instvar vagent_array_;
    $self instvar subgraph_list_;
    $self instvar subgraph_info_;
    $self instvar id_;
    $self instvar comm_obj_;
    $self instvar callback_array_;
    $self instvar next_subgraph_;
    $self instvar sid_;
    $self instvar ctoken_count_;
    $self instvar token_rate_period_;
    $self instvar rate_pressure_ rate_constant_;

    set rate_pressure_ 1.25;
#    set rate_constant_ [expr 1.0/1000.0];
    set rate_constant_ 0.0;

    set input_id_list_ "";
    set id_ $id;
    set comm_obj_ [new AnnounceListenManager/GraphComm [$self get_option mtu] $control_spec];
    $comm_obj_ register $id "$self recv_graph_comm ";

    set sid_ -1;
    set subgraph_list_ "";
    set next_subgraph_ "";
    set ctoken_count_ 0;
    set token_rate_period_ 1000;

    $self instvar trigger_count_;
    $self instvar fire_count_;

    set trigger_count_ 0;
    set fire_count_ 0;
}

TemporalSelector instproc send_to_subgraph {sid rid mesg {repeat_period 0} {deliver_once_flag 0}} {
    $self instvar subgraph_info_;
    $self instvar id_;

    set comm_obj $subgraph_info_($sid,comm_obj);
    $comm_obj do_announce $id_ $rid $mesg $repeat_period $deliver_once_flag;
}

TemporalSelector instproc decr_ctoken_count {} {
    $self instvar ctoken_count_;

    incr ctoken_count_ -1;
}

TemporalSelector instproc handle_trigger_count {} {
    $self instvar trigger_count_;
    $self instvar token_rate_period_;

    incr trigger_count_;
    after $token_rate_period_ "$self set trigger_count_ \[expr \[$self set trigger_count_\] - 1\]";
}

TemporalSelector instproc handle_fire_count {} {
    $self instvar fire_count_;
    $self instvar token_rate_period_;

    incr fire_count_;
    after $token_rate_period_ "$self set fire_count_ \[expr \[$self set fire_count_\] - 1\]";
}

TemporalSelector instproc recv_subgraph_comm {sid rid mesg} {
    $self instvar callback_array_;

    if {[info exists callback_array_($sid,$rid)]} {
	set cb $callback_array_($sid,$rid);
	unset callback_array_($sid,$rid);
	eval $cb [list $mesg];
    }

    if {[lindex $mesg 0] == "trigger_completion_token"} {
	$self instvar next_subgraph_;
	$self instvar subgraph_list_;

	$self instvar ctoken_count_ token_rate_period_;

	incr ctoken_count_;
	after $token_rate_period_ "$self decr_ctoken_count";

	set idx [lsearch $next_subgraph_ $sid];
	if {$idx != -1} {
	    set next_subgraph_ [lreplace $next_subgraph_ $idx $idx];
	}
 
	if {[lsearch $subgraph_list_ $sid] != -1} {
	    set next_subgraph_ [linsert $next_subgraph_ 0 $sid];
	}
    }
}

TemporalSelector instproc SetupInputs {sid req_id reply} {
    $self instvar subgraph_list_;
    $self instvar subgraph_info_;
    $self instvar vagent_array_;

    set comm_obj $subgraph_info_($sid,comm_obj);
    $comm_obj stop_resend $req_id;
    
    set in_list [lindex $reply 1];
    
    $self instvar input_id_list_;
    $self instvar input_info_;
    
    set input_id_list_ $in_list;

    foreach i $input_id_list_ {
	set input_info_($i,spec) "";
	set input_info_($i,trigger) 0;
	set input_info_($i,selector) "";
    }
}

TemporalSelector instproc recv_graph_comm {rid mesg} {
    eval $self instvar [$self info vars];

    set cmd [lindex $mesg 0];

    # Don't do anything if we don't have any subgraphs unless cmd
    # is add_subgraph
    # XXX This might be a problem since underlying comm object
    #     thinks we've handled this message. Maybe we need
    #     a way to tell comm object not to assume we've handled
    #     the message, or queue messages that we get when we weren't
    #     ready and deal with them when we are?
    if {$subgraph_list_ == ""} {
	if {$cmd != "add_subgraph"} {
	    return;
	}
    }


    switch -exact -- $cmd {
	query_input_ids {
	    $self send_graph_comm $rid [list input_ids $input_id_list_];
	}
	query_input_info {
	    set in_id [lindex $mesg 1];
	    if {[lsearch $input_id_list_ $in_id] != -1} {
		$self send_graph_comm $rid [list input_info $in_id [list \
			[list spec $input_info_($in_id,spec)] \
			[list trigger $input_info_($in_id,trigger)]]];
	    } else {
		$self send_graph_comm $rid [list input_info $in_id [list \
			[list error "No such input_id $in_id"]]];
	    }
	}
	query_parameter_ids {
	    puts "Doing query_parameter_ids"
	    set sub [lindex $subgraph_list_ 0];
	    set mid [$self send_to_subgraph $sub -1 "query_parameter_ids" 100];
	    set callback_array_($sub,$mid) "$self forward_response $rid $sub $mid "
	}
	query_parameter_info {
	    set pid [lindex $mesg 1];
	    set sub [lindex $subgraph_list_ 0];
	    set mid [$self send_to_subgraph $sub -1 "query_parameter_info $pid" 100];
	    set callback_array_($sub,$mid) "$self forward_response $rid $sub $mid "
	}
	set_parameter {
	    set pid [lindex $mesg 1];
	    set pval [lindex $mesg 2];
	    
	    foreach s $subgraph_list_ {
		$self send_to_subgraph $s -1 [list set_parameter $pid $pval];
	    }
	}
	set_input_spec {
	    set in_id [lindex $mesg 1];
	    set spec [lindex $mesg 2];
	    if {[lsearch $input_id_list_ $in_id] != -1} {
		$self set_input_spec $in_id $spec;
	    }
	}
	trigger {
	    $self trigger;
	}
	set_input_trigger {
	    set in_id [lindex $mesg 1];
	    set trigger_flag [lindex $mesg 2];
	    if {[lsearch $input_id_list_ $in_id] != -1} {
		if {$input_info_($in_id,trigger) != $trigger_flag} {
		    set input_info_($in_id,trigger) $trigger_flag;
		
		    if {$input_info_($in_id,selector) != ""} {
			if {$input_info_($in_id,trigger) != 0} {
			    $input_info_($in_id,selector) set_callback "$self trigger; # ";
			} else {
			    $input_info_($in_id,selector) set_callback "";
			}
		    }
		}
	    }
	    set some_trigger 0;
	    foreach in_id $input_id_list_ {
		if {$input_info_($in_id,trigger) != 0} {
		    set some_trigger 1;
		    break;
		}
	    }
	    if {$some_trigger == 0} {
		$self instvar last_time_click_;
		unset last_time_click_;
	    }
	}
	add_subgraph {
	    $self instvar subgraph_info_;
	    $self instvar callback_array_;

	    set new_subgraph [lindex $mesg 1];

	    foreach sid $subgraph_list_ {
		if {$subgraph_info_($sid,spec) == $new_subgraph} {
		    return;
		}
	    }

	    incr sid_;
	    lappend subgraph_list_ $sid_;
	    lappend next_subgraph_ $sid_;

	    set subgraph_info_($sid_,comm_obj) [new AnnounceListenManager/GraphComm [$self get_option mtu] $new_subgraph];
	    $subgraph_info_($sid_,comm_obj) register $id_ "$self recv_subgraph_comm $sid_ ";
	    set subgraph_info_($sid_,spec) $new_subgraph;

	    if {[llength $subgraph_list_] == 1} {
		set input_id_request [$self send_to_subgraph $sid_ -1 "query_input_ids" 100];
		set callback_array_($sid_,$input_id_request) "$self SetupInputs $sid_ $input_id_request ";
	    } else {		
		foreach i $input_id_list_ {
		    if {$input_info_($i,selector) != ""} {
			$input_info_($i,selector) add_subgraph $sid_;
		    }
		}
	    }
	}
	del_subgraph {
	    set spec_to_del [lindex $mesg 1];
	    set sid_to_del "";
	    foreach s $subgraph_list_ {
		if {$subgraph_info_($s,spec) == $spec_to_del} {
		    set sid_to_del $s;
		}
	    }
	    if {$sid_to_del == ""} {
		return;
	    }
	    set idx [lindex $subgraph_list_ $sid_to_del];
	    set subgraph_list_ [lreplace $subgraph_list_ $idx $idx];

	    set idx [lindex $next_subgraph_ $sid_to_del];
	    if {$idx != -1} {
		set next_subgraph_ [lreplace $next_subgraph_ $idx $idx];
	    }

	    foreach i $input_id_list_ {
		set selector $input_info_($i,selector);
		if {$selector != ""} {
		    $selector del_subgraph $sid_to_del;
		}
	    }
	    delete $subgraph_info_($sid_to_del,comm_obj);
	    unset subgraph_info_($sid_to_del,comm_obj);
	    unset subgraph_info_($sid_to_del,spec);

	    if {$subgraph_list_ == ""} {
		# All the subgraphs are gone.
		# XXX Need to handle this case.
	    }
	}
    }
}

TemporalSelector instproc trigger {} {
    $self instvar next_subgraph_;
    $self instvar subgraph_list_;
    $self instvar subgraph_info_;
    $self instvar input_id_list_;
    $self instvar input_info_;

    $self instvar trigger_count_;
    $self instvar fire_count_;

    $self handle_trigger_count;
    # This is a total hack to prevent overflowing the processors.
    
    $self instvar last_time_click_;
    $self instvar update_status;
    $self instvar ctoken_count_ token_rate_period_;
    $self instvar rate_pressure_ rate_constant_;

    if {![info exists last_time_click_]} {
	set last_time_click_ [clock clicks];
    }
    
    set now [clock clicks];

    set target_rate [expr ($rate_pressure_ * \
	    (($ctoken_count_*1.0) / ($token_rate_period_*1.0)) + \
	    $rate_constant_)];
    if {$target_rate <= 0.0} {
	set target_rate 1.0;
    }

    set target_interval [expr  1.0 / $target_rate];
    
    set tdiff [expr ($now - $last_time_click_) / 1000.0];

    if {$tdiff < $target_interval} {
	set update_status 0;
    } else {
	set update_status 1;
	set last_time_click_ [expr $last_time_click_ + int(($target_interval * 1000.0))];
    }

    set s "";
    if {$update_status} {
	# set next_subgraph_ [expr ($next_subgraph_ + 1) % $num_subgraphs];

	set s [lindex $next_subgraph_ 0];
	set next_subgraph_ [lreplace $next_subgraph_ 0 0];
	lappend next_subgraph_ $s;

	# Build and send trigger vector
	$self instvar tvector_;
	set tvector_ "";
	foreach i $input_id_list_ {
	    set selector $input_info_($i,selector);
	    if {$selector != ""} {
		set ts [$selector get_current_ts];
		lappend tvector_ [list $i $ts];
	    }
	}
	$self send_to_subgraph $s -1 [list trigger_vector $tvector_];
	$self handle_fire_count;
    }
    foreach i $input_id_list_ {
	set selector $input_info_($i,selector);
	if {$selector != ""} {
	    $selector update_subgraph $s
	}
    }
}

TemporalSelector instproc set_input_spec {in_id spec} {
    $self instvar input_info_;
    $self instvar vagent_array_;
    $self instvar subgraph_list_;
    $self instvar subgraph_info_;

    if {$input_info_($in_id,spec) == $spec} {
	return;
    }

    # If input already associated with a spec, delete selector no longer
    # needed.

    if {$input_info_($in_id,spec) != ""} {
	set selector $input_info_($in_id,selector);
	if {$selector != ""} {
	    set vagent [$selector set vagent_];
	    $vagent delete_decoder [$selector get_decoder];
	    delete $selector;
	    set input_info_($id,selector) "";
	}
	set input_info_($in_id,spec) "";
    }

    set spec_split [split $spec "/"];

    set addr [lindex $spec_split 0];
    set port [lindex $spec_split 1];
    set srcid [lindex $spec_split 2];

    if {![info exists vagent_array_($addr,$port)]} {
	set vagent_array_($addr,$port) [new PsvpVideoAgent $addr/$port];
    }
    set vagent $vagent_array_($addr,$port);

    set src [$vagent get_source_by_id $srcid];

    if {$src == ""} {
	$vagent set_create_decoder_callback $srcid "$self set_input_spec_cb $in_id $spec"
    } else {
	set fmt_name [$vagent classmap [$src format_name]];

	set selector [new TemporalSelectorInput/${fmt_name} $self $in_id]

	if {$selector == ""} {
	    set selector [new TemporalSelectorInput/Null $self $in_id];
	} 
	if {$input_info_($in_id,trigger) != 0} {
	    $selector set_callback "$self trigger; # ";
	}
	foreach s $subgraph_list_ {
	    $selector add_subgraph $s;
	}
	$vagent set_src_decoder $src [$selector get_decoder];
	
	$selector set vagent_ $vagent;

	set input_info_($in_id,selector) $selector;
    }
    set input_info_($in_id,spec) $spec;
}

TemporalSelector instproc set_input_spec_cb {id spec src} {
    $self instvar input_info_
    $self instvar vagent_array_
    $self instvar subgraph_info_
    $self instvar subgraph_list_
    
    if {$input_info_($id,spec) != $spec} {
	puts "$input_info_($id,spec) != $spec !!!";
	return;
    }

    if {$input_info_($id,selector) != ""} {
	puts "$input_info_($id,selector) != {} !!!";
	return;
    }

    # Find videoagent and source
    
    set spec_split [split $spec "/"];
    
    set addr [lindex $spec_split 0];
    set port [lindex $spec_split 1];
    set srcid [lindex $spec_split 2];
    
    set vagent $vagent_array_($addr,$port);
    
    # Setup decoder.
    
    set fmt_name [$vagent classmap [$src format_name]];
    
    set selector [new TemporalSelectorInput/${fmt_name} $self $id];

    if {$selector == ""} {
	# No such selector. Create a NULL decoder and associate that.
	set selector [new TemporalSelectorInput/Null $self $id];
    } else {
	if {$input_info_($id,trigger) != 0} {
	    $selector set_callback "$self trigger; #";
	}
	foreach s $subgraph_list_ {
	    $selector add_subgraph $s;
	}
    }
    # Update input info array

    $selector set vagent_ $vagent;
    
    set input_info_($id,selector) $selector;
    return [$selector get_decoder];
}


TemporalSelector instproc forward_response {orid sub mid mesg} {
    $self instvar subgraph_info_;

    set comm_obj $subgraph_info_($sub,comm_obj);
    $comm_obj stop_resend $mid;

    $self send_graph_comm $orid $mesg;
}

TemporalSelector instproc send_graph_comm {rid mesg} {
    $self instvar comm_obj_;
    $self instvar id_;
    
    $comm_obj_ do_announce $id_ $rid $mesg;
}

TemporalSelector instproc GenerateSpec {} {
    # OK, this is a REAL hack
    
    global spec_generator;
    
    if {![info exists spec_generator(init)]} {
	set spec_generator(init) 1;
	set spec_generator(b1) 224;
	set spec_generator(b2) 5;
	set spec_generator(b3) 6;
	set spec_generator(b4) 1;
	set spec_generator(port) 22334;
    }
    
    incr spec_generator(b4);

    return "$spec_generator(b1).$spec_generator(b2).$spec_generator(b3).$spec_generator(b4)/$spec_generator(port)/16";
}