#
# Copyright (c) 1998 Regents of the University of California.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#       This product includes software developed by the Computer Systems
#       Engineering Group at Lawrence Berkeley Laboratory.
# 4. Neither the name of the University nor of the Laboratory may be used
#    to endorse or promote products derived from this software without
#    specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# ------------
#
# Filename: pane-manager.tcl
#   -- Created on Wed Apr 15 1998
#   -- Author: Cynthia Romer <cromer@cs.berkeley.edu>
#
#  @(#) $Header: /usr/src/mash/repository/mash/mash-1/tcl/ui_tools/pane-manager.tcl,v 1.1 1998/04/21 10:31:56 cromer Exp $
#

#
# Sets up vertical or horizontal panes.  Two frames, <i>f1</i> and
# <i>f2</i>, are placed inside another frame, which represents a grip
# that is used to adjust the boundary between the two panes.  The
# arguments, which are optional, are "-orient [vertical|horizontal]",
# "-percent [0-1]", and "-in [parent window]".
# This was copied almost directly from example 22-5 of "Practical Programming in Tcl and Tk" -Welch
#
Class PaneManager

#
# The code is set to work with either horizontal or vertical layouts.
# The pane(D) variable is either <i>x</i>, for a horizontal layout, or
# <i>y</i>, for a vertical layout.  Remember the previous position in
# pane(lastD) and use that to update the percentage split between the
# two contained panes.
#
PaneManager instproc pane_drag {master D} {
	$self instvar pane
	
	upvar #0 pane_$master pane
	if [info exists pane(lastD)] {
		set delta [expr double($pane(lastD) - $D) \
				/ $pane(size)]
		set pane(-percent) [expr $pane(-percent) - $delta]
		if {$pane(-percent) < 0.0} {
			set pane(-percent) 0.0
		} elseif {$pane(-percent) > 1.0} {
			set pane(-percent) 1.0
		}
		$self pane_geometry $master
	}
	set pane(lastD) $D
}

#
PaneManager instproc pane_stop {master} {
	$self instvar pane
	
	upvar #0 pane_$master pane
	catch {unset pane(lastD)}
}

#
# Adjust the positions of the frames.  Called when the main window is
# resized, so it updates pane(size).  Also called as the user drags the
# grip.  For a vertical layout, the grip is moved by setting its
# relative Y position.  The size of the two contained frames is set with
# a relative height.  This is combined wit the fixed height of -1 to get
# some space between the two frames.
#
PaneManager instproc pane_geometry {master} {
	$self instvar pane
	
	upvar #0 pane_$master pane
	if {$pane(D) == "X"} {
		place $pane(1) -relwidth $pane(-percent)
		place $pane(2) -relwidth [expr 1.0 - $pane(-percent)]
		place $pane(grip) -relx $pane(-percent)
		set pane(size) [winfo width $master]
	} else {
		place $pane(1) -relheight $pane(-percent)
		place $pane(2) -relheight [expr 1.0 - $pane(-percent)]
		place $pane(grip) -rely $pane(-percent)
		set pane(size) [winfo height $master]
	}
}

#
# Try this to see how PaneManager works.
#
proc pane_test {{p .p} {orient vert}} {
	catch {destroy $p}
	frame $p -width 200 -height 200
	label $p.1 -bg blue -text foo
	label $p.2 -bg green -text bar
	pack $p -expand true -fill both
	pack propagate $p off
	new PaneManager $p.1 $p.2 -in $p -orient $orient -percent 0.3
}

#
# Given two widgets and an optional set of parameters, sets them up as
# vertical/horizontal panes offering the user a grip with which to
# adjust the boundary between them.
#
PaneManager instproc init {f1 f2 args} {
	$self instvar pane
	
	# Map optional arguments into array values
	set t(-orient) vertical
	set t(-percent) 0.5
	set t(-in) [winfo parent $f1]
	array set t $args

	# Keep state in an array associated with the master frame
	set master $t(-in)
	upvar #0 pane_$master pane
	array set pane [array get t]

	# Create the grip and set placement attributes that
	# will not change.  A thin divider line is achieved by
	# making the two frames one pixel smaller in the
	# adjustable dimension and making the main frame black.

	set pane(1) $f1
	set pane(2) $f2
	set pane(grip) [frame $master.grip -background gray50 \
			-width 10 -height 10 -bd 1 -relief raised \
			-cursor crosshair]
	if {[string match vert* $pane(-orient)]} {
		set pane(D) Y; # Adjust boundary in Y direction.
		place $pane(1) -in $master -x 0 -rely 0.0 -anchor nw \
				-relwidth 1.0 -height -1
		place $pane(2) -in $master -x 0 -rely 1.0 -anchor sw \
				-relwidth 1.0 -height -1
		place $pane(grip) -in $master -anchor c -relx 0.8
	} else {
		set pane(D) X; # Adjust boundary in X direction
		place $pane(1) -in $master -relx 0.0 -y 0 -anchor nw \
				-relheight 1.0 -width -1
		place $pane(2) -in $master -relx 1.0 -y 0 -anchor ne \
				-relheight 1.0 -width -1
		place $pane(grip) -in $master -anchor c -rely 0.8
	}
	$master configure -background black

	# Set up bindings for resize, <Configure>, and
	# for dragging the grip.

	# The <Configure> event occurs when the containing frame is resized by the user.
	bind $master <Configure> [list $self pane_geometry $master]
	# When the user presses the mouse button over the grip and drags it,
	# there is a <ButtonPress-1> event, one or more <B1-Motion> events, and finally
	# a <ButtonRelease-1> event.
	bind $pane(grip) <ButtonPress-1> \
			[list $self pane_drag $master %$pane(D)]
	bind $pane(grip) <B1-Motion> \
			[list $self pane_drag $master %$pane(D)]
	bind $pane(grip) <ButtonRelease-1> \
			[list $self pane_stop $master]

	# Do the initial layout

	$self pane_geometry $master
}



